// Copyright (C) 2011-2013 by Maximilian Hönig
// This file is part of "DunGen - the Dungeongenerator".
// For conditions of distribution and use, see licence.txt provided together with DunGen.

#include "Corridor.h"
#include "DockingSite.h"
#include "Helperfunctions.h"

// ======================================================
// computation of t, position and derivation
// ======================================================

inline irr::core::vector3d<double> DunGen::CCorridor::ComputePosition(double t_)
{
	return (t_*t_*t_* PositionCoefficients[0] + t_*t_* PositionCoefficients[1] + t_* PositionCoefficients[2] + PositionCoefficients[3]);
}

inline irr::core::vector3d<double> DunGen::CCorridor::ComputeDerivation(double t_)
{
	return (t_*t_* DerivationCoefficients[0] + t_* DerivationCoefficients[1] + DerivationCoefficients[2]);
}

inline irr::core::vector2d<double> DunGen::CCorridor::ComputeT(
	const irr::core::vector3d<double>& lastPosition_, double lastT_, double distance_)
{
	// computation of next t with bisection

	// set start values
	double t = 1.0;
	irr::core::vector3d<double> newPosition = ComputePosition(t);
	double actDistanceSQ = (newPosition-lastPosition_).getLengthSQ(); // used squared distances for faster computation
	double distanceSQ = distance_*distance_;
	
	// if distance to t=1.0 (endpoint) already too small, no desired distance possible
	if (actDistanceSQ <= distanceSQ)
		return irr::core::vector2d<double>(t,sqrt(actDistanceSQ));

	// interval boundaries for searching t
	double upperBorder = 1.0;
	double lowerBorder = lastT_;	
	
	// iterate until desired precision is achieved
	while (abs(upperBorder-lowerBorder)>Precision)
	{
		// distance too big -> lower the distance, use lower interval half
		if (actDistanceSQ>distanceSQ)
			upperBorder = t;
		// distance too small -> enlarge distance, use upper interval half
		else
			lowerBorder = t;
		
		// update values:
		// t is median of upper and lower boundary
		t = 0.5*(lowerBorder+upperBorder);
		
		// compute new position and distance
		newPosition = ComputePosition(t);
		actDistanceSQ = (newPosition-lastPosition_).getLengthSQ();
	}

	// return result
	return irr::core::vector2d<double>(t,distance_);
}

// ======================================================
// helper functions
// ======================================================

inline double DunGen::CCorridor::ComputePositionInterpolationFactor(
	const irr::core::vector2d<double>& ma_, const irr::core::vector2d<double>& mb_, double angleAMP_)
{
	double ma = ma_.getLength();
	double mb = mb_.getLength();
	irr::core::vector2d<double> vecAB = mb_-ma_;
	double ab = (vecAB).getLength();

	if (ma == 0.0 || ab == 0.0)
		return 0.0;

	// compute angle by law of cosines: 0 to PI
	double angleBAM = acos(-ma_.dotProduct(vecAB) / (ma * ab));
	// law of internal angles of a triangle
	double angleMPA = M_PI - angleBAM - angleAMP_;
	// law of sines
	double ap = sin(angleAMP_)/sin(angleMPA) * ma;

	return ap/ab;
}

inline irr::core::vector2d<double> DunGen::CCorridor::ProjectVectorToPlane(
	const irr::core::vector3d<double>& vector_, const irr::core::vector3d<double>& left_, const irr::core::vector3d<double>& up_)
{
	return irr::core::vector2d<double>(-vector_.dotProduct(left_),vector_.dotProduct(up_));
}

// ======================================================
// creation of the corridor
// ======================================================

double DunGen::CCorridor::CreateCorridor(const SCorridorProfile& profile_,
	const SDockingSite& dockingSite0_, const SDockingSite& dockingSite1_, double distance_, double distanceTextureYPerDistance1_)
{
	// spline parameter t
	double t;
	// position and derivation for current t value
	irr::core::vector3d<double> actPosition;
	irr::core::vector3d<double> actDerivation;
	// local coordinate system (left handed)
	irr::core::vector3d<double> front;
	irr::core::vector3d<double> left;
	irr::core::vector3d<double> up;

	// parameters to determine the region of interests for the visibility test
	// they should have a high distance to axis P(1)-P(0) -> local maxima (there should be exactly 1 or 2)
	irr::core::vector3d<double> roiAxis = (Position[1]-Position[0]).normalize();
	SViewRegionOfInterest actRegionOfInterest;
	double actDistanceSQ = 0.0;
	double lastDistanceSQ;
	bool roiStored = false;

	// security limit for meshbuffer: one profile segment
	unsigned int resMaxNumVertices = MaxVertexCount - profile_.Point.size();
	// create new mesh
	MeshCorridor = new irr::scene::SMesh();

	// create and add first buffer
	irr::scene::SMeshBuffer* meshBuffer = new irr::scene::SMeshBuffer();
	MeshCorridor->addMeshBuffer(meshBuffer);
	meshBuffer->drop(); // can be dropped now, because it is hold by the mesh
	// reserve for worst case
	meshBuffer->Vertices.set_used(MaxVertexCount); 
	meshBuffer->Indices.set_used(6*MaxVertexCount);	// 6 indices per vertex (2 triangles per vertex)
	unsigned int bufferVertices = 0;
	unsigned int bufferIndizes = 0;

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// add fist corridor segment
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	double actTextureCoordY = 0;
	irr::core::vector3d<double> tmp;

	t = 0.0;
	actPosition = Position[0];
	actDerivation = Derivation[0];
	// compute coordinate system (left handed)
	front = actDerivation.normalize();
	left = front.crossProduct(UpStandard);
	left.normalize();
	up = left.crossProduct(front);

	// for every point of the profile:
	for (unsigned int i = 0; i<profile_.Point.size(); ++i)
	{
		// compute new vertex and add it to buffer, resulting position: (-left,up)*(X,Y)
		tmp = actPosition - profile_.Point[i].X*left + profile_.Point[i].Y*up;
		irr::video::S3DVertex& v = meshBuffer->Vertices[bufferVertices++];
		v.Pos.set(vec3F(tmp));
		
		// compute normal
		tmp = actPosition - tmp;
		tmp.normalize();
		v.Normal.set(vec3F(tmp));

		// set texture coordinates
		v.TCoords.set(static_cast<irr::f32>(profile_.TextureX[i]),static_cast<irr::f32>(actTextureCoordY));
	}

	// first segment is also a region of interest
	actRegionOfInterest.Position = actPosition;
	actRegionOfInterest.Left = left;
	actRegionOfInterest.Up = up;
	ViewTestRegions.push_back(actRegionOfInterest);

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// go on further on the spline
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	irr::core::vector2d<double> computeTResult;
	while (t < 1.0)
	{
		// actualize t
		computeTResult = ComputeT(actPosition, t, distance_);
		t = computeTResult.X;
		// compute y texture coordinate and position
		actTextureCoordY+= computeTResult.Y * distanceTextureYPerDistance1_;
		actPosition = ComputePosition(t);

		// compute coordinate system (left handed)
		actDerivation = ComputeDerivation(t);
		front = actDerivation.normalize();
		left = front.crossProduct(UpStandard);
		left.normalize();
		up = left.crossProduct(front);

		// look for region of interest
		lastDistanceSQ = actDistanceSQ;
		tmp = actPosition-Position[0];
		actDistanceSQ = (tmp - tmp.dotProduct(roiAxis)*roiAxis).getLengthSQ();
		// distance gets larger: store data
		if (actDistanceSQ>lastDistanceSQ)
		{
			actRegionOfInterest.Position = actPosition;
			actRegionOfInterest.Left = left;
			actRegionOfInterest.Up = up;
			roiStored = false;
		}
		// distance gets smaller -> local maximum -> store region of interest
		else if (!roiStored)
		{
			// data of the last pass is stored (the data of the local maximum)
			ViewTestRegions.push_back(actRegionOfInterest);
			roiStored = true;
		}

		// for every point of the profile:
		for (unsigned int i = 0; i<profile_.Point.size(); ++i)
		{
			// compute new vertex and add it to buffer, resulting position: (-left,up)*(X,Y)
			tmp = actPosition - profile_.Point[i].X*left + profile_.Point[i].Y*up;
			irr::video::S3DVertex& v = meshBuffer->Vertices[bufferVertices++];
			v.Pos.set(vec3F(tmp));
		
			// compute normal
			tmp = actPosition - tmp;
			tmp.normalize();
			v.Normal.set(vec3F(tmp));

			// set texture coordinates
			v.TCoords.set(static_cast<irr::f32>(profile_.TextureX[i]),static_cast<irr::f32>(actTextureCoordY));
		}
		
		// create triangles
		// indices counter clockwise from outside
		// triangle 1
		meshBuffer->Indices[bufferIndizes++] = bufferVertices - profile_.Point.size() - 1;	// last point of last profile 
		meshBuffer->Indices[bufferIndizes++] = bufferVertices - profile_.Point.size();		// first point of actual profile
		meshBuffer->Indices[bufferIndizes++] = bufferVertices - 1;							// last point of actual profile

		// triangle 1
		meshBuffer->Indices[bufferIndizes++] = bufferVertices - profile_.Point.size() - 1;	// last point of last profile
		meshBuffer->Indices[bufferIndizes++] = bufferVertices - 2*profile_.Point.size();	// first point of last profile
		meshBuffer->Indices[bufferIndizes++] = bufferVertices - profile_.Point.size();		// first point of actual profile
	
		// rest:
		for (unsigned int i = 1; i<profile_.Point.size(); i++)
		{
			// triangle 1
			meshBuffer->Indices[bufferIndizes++] = bufferVertices - 2*profile_.Point.size() + i - 1;
			meshBuffer->Indices[bufferIndizes++] = bufferVertices - profile_.Point.size() + i;
			meshBuffer->Indices[bufferIndizes++] = bufferVertices - profile_.Point.size() + i - 1;			

			// triangle 2
			meshBuffer->Indices[bufferIndizes++] = bufferVertices - 2*profile_.Point.size() + i - 1;
			meshBuffer->Indices[bufferIndizes++] = bufferVertices - 2*profile_.Point.size() + i;
			meshBuffer->Indices[bufferIndizes++] = bufferVertices - profile_.Point.size() + i;			
		}

		// if meshbuffer full:
		if (bufferVertices > resMaxNumVertices && t < 1.0)
		{
			// old buffer: set final size and bounding box
			meshBuffer->Vertices.reallocate(bufferVertices);
			meshBuffer->Indices.reallocate(bufferIndizes);
			meshBuffer->recalculateBoundingBox();

			// create and add new buffer
			meshBuffer = new irr::scene::SMeshBuffer();
			bufferVertices = 0;	
			bufferIndizes = 0;
			MeshCorridor->addMeshBuffer(meshBuffer);
			meshBuffer->drop();
			meshBuffer->Vertices.set_used(MaxVertexCount);
			meshBuffer->Indices.set_used(6*MaxVertexCount);

			// add vertices of the last segment again, so the mesh remains connected
			// for every point of the profile:
			for (unsigned int i = 0; i<profile_.Point.size(); ++i)
			{
				// compute new vertex and add it to buffer, resulting position: (-left,up)*(X,Y)
				tmp = actPosition - profile_.Point[i].X*left + profile_.Point[i].Y*up;
				irr::video::S3DVertex& v = meshBuffer->Vertices[bufferVertices++];
				v.Pos.set(vec3F(tmp));
		
				// compute normal
				tmp = actPosition - tmp;
				tmp.normalize();
				v.Normal.set(vec3F(tmp));

				// set texture coordinates
				v.TCoords.set(static_cast<irr::f32>(profile_.TextureX[i]),static_cast<irr::f32>(actTextureCoordY));
			}
		}
	} // END: while

	// buffer: set final size and bounding box
	meshBuffer->Vertices.reallocate(bufferVertices);
	meshBuffer->Indices.reallocate(bufferIndizes);
	meshBuffer->recalculateBoundingBox();

	// set final bounding box of the mesh
	MeshCorridor->recalculateBoundingBox();	

	// last segment is also a region of interest
	actRegionOfInterest.Position = actPosition;
	actRegionOfInterest.Left = left;
	actRegionOfInterest.Up = up;
	ViewTestRegions.push_back(actRegionOfInterest);

	// return the y texture coordinate of the last segment (the first segment has 0.0)
	return actTextureCoordY;
}

// ======================================================
// creation of the adapters
// ======================================================

void DunGen::CCorridor::CreateAdapter(unsigned int end_,
	const SDockingSite& dockingSite_, const SCorridorProfile& profile_, double lastTextureY_, double distanceTextureYPerDistance1_)
{
	// create new mesh
	MeshAdapter[end_] = new irr::scene::SMesh();
	
	// if the docking site has no points -> no docking site should be created
	if (0 == dockingSite_.Point.size())
	{
		MeshAdapter[end_]->recalculateBoundingBox();
		return;
	}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// prepare corridor segment
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	
	// compute position, derivation and local coordinate system
	irr::core::vector3d<double> actPosition = Position[end_];
	irr::core::vector3d<double> actDerivation = Derivation[end_];
	irr::core::vector3d<double> front = actDerivation.normalize();
	irr::core::vector3d<double> left = front.crossProduct(UpStandard);
	left.normalize();
	irr::core::vector3d<double>	up = left.crossProduct(front);

	double angle1, angle2;
	unsigned int startIndex, actIndex;

	// compute polar angles for each point: clockwise, 0 degrees is left-vector
	std::vector<double> angles;
	for (unsigned int i = 0; i<profile_.Point.size(); ++i)
	{
		// atan2: -PI bis PI
		angle1 = atan2( profile_.Point[i].Y,- profile_.Point[i].X);
		angles.push_back(angle1);
	}
		
	// sort by increasing angle:

	// compute start index
	angle1 = 3*M_PI;
	for (unsigned int i = 0; i<angles.size(); ++i)
		if (angles[i]<angle1)
		{
			angle1 = angles[i];
			startIndex = i;
		};

	// now sort
	SCorridorProfile profileCorridorSorted;
	std::vector<double> angleCorridorSorted;
	
	// first value
	actIndex = startIndex;
	profileCorridorSorted.Point.push_back(profile_.Point[actIndex]);
	profileCorridorSorted.TextureX.push_back(profile_.TextureX[actIndex]);
	angleCorridorSorted.push_back(angles[actIndex]);
	++actIndex;
	if (actIndex==profile_.Point.size())
		actIndex = 0;
	
	// remaining values
	while (actIndex != startIndex)
	{
		profileCorridorSorted.Point.push_back(profile_.Point[actIndex]);
		profileCorridorSorted.TextureX.push_back(profile_.TextureX[actIndex]);
		angleCorridorSorted.push_back(angles[actIndex]);
		++actIndex;
		if (actIndex==profile_.Point.size())
			actIndex = 0;
	}

	// compute 3d points of the segment
	std::vector<irr::core::vector3d<double>> corridorVertices;
	irr::core::vector3d<double> tmp;
	// for every point of the profile:
	for (unsigned int i = 0; i<profileCorridorSorted.Point.size(); ++i)
	{
		// resulting position: (-left,up)*(X,Y)
		tmp = actPosition - profileCorridorSorted.Point[i].X*left + profileCorridorSorted.Point[i].Y*up;
		corridorVertices.push_back(tmp);
	}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// prepare docking site
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	// compute local coordinate system
	irr::core::vector3d<double> leftOfDS = dockingSite_.Normal.crossProduct(UpStandard);
	leftOfDS.normalize();
	irr::core::vector3d<double> upOfDS = leftOfDS.crossProduct(dockingSite_.Normal);
	
	// compute polar angles for each point: clockwise, 0 degrees is left-vector
	angles.clear();
	for (unsigned int i = 0; i<dockingSite_.Point.size(); ++i)
	{
		// origin is the center
		tmp = dockingSite_.Point[i]-dockingSite_.Center;		
		// use projected 2D-coordinates
		angle1 = atan2(tmp.dotProduct(upOfDS),tmp.dotProduct(leftOfDS));
		angles.push_back(angle1);
	}

	// sort by increasing angle:

	// compute start index
	angle1 = 3*M_PI;
	for (unsigned int i = 0; i<angles.size(); ++i)
		if (angles[i]<angle1)
		{
			angle1 = angles[i];
			startIndex = i;
		};

	// now sort
	SDockingSite dockingSiteSorted;
	std::vector<double> anglesDockingSiteSorted;
	
	// first value
	actIndex = startIndex;
	dockingSiteSorted.Point.push_back(dockingSite_.Point[actIndex]);
	anglesDockingSiteSorted.push_back(angles[actIndex]);
	++actIndex;
	if (actIndex==dockingSite_.Point.size())
		actIndex = 0;
	
	// remaining values
	while (actIndex != startIndex)
	{
		dockingSiteSorted.Point.push_back(dockingSite_.Point[actIndex]);
		anglesDockingSiteSorted.push_back(angles[actIndex]);
		++actIndex;
		if (actIndex==dockingSite_.Point.size())
			actIndex = 0;
	}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// create drawing instructions of the adapter
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	std::vector<SAdapterDrawingInstruction> adapterDrawingInstructions;

	unsigned int actIndexDockingSite = 0;
	unsigned int actIndexCorridorSegment = 0;
	unsigned int lastIndexDockingSite = dockingSiteSorted.Point.size()-1;
	unsigned int lastIndexCorridorSegment = profileCorridorSorted.Point.size()-1;
	double interpolationFactor;

	bool closedDockingSite = false;
	bool closedCorridorSegment = false;

	// walk over all vertices (adapter has to be closed)
	while(!(closedDockingSite&&closedCorridorSegment))
	{
		// which is the next vertex? the one with the smaller angle!
		if (
			closedDockingSite ||
			((angleCorridorSorted[actIndexCorridorSegment]<anglesDockingSiteSorted[actIndexDockingSite]) && !closedCorridorSegment)	
			)
		// corridor vertex
		{
			// compute interpolation factor
			angle1 = angleCorridorSorted[actIndexCorridorSegment] - anglesDockingSiteSorted[lastIndexDockingSite];
			
			interpolationFactor = ComputePositionInterpolationFactor(
				//  vector (p-m) in projected coordinates
				ProjectVectorToPlane( dockingSiteSorted.Point[lastIndexDockingSite]-dockingSite_.Center, leftOfDS, upOfDS),
				ProjectVectorToPlane( dockingSiteSorted.Point[actIndexDockingSite]-dockingSite_.Center, leftOfDS, upOfDS),
				angle1);
			
			// compute coordinates
			tmp = (1-interpolationFactor) * dockingSiteSorted.Point[lastIndexDockingSite]
						+ interpolationFactor * dockingSiteSorted.Point[actIndexDockingSite];

			// create new drawing instruction
			SAdapterDrawingInstruction newDrawingInstruction;
			newDrawingInstruction.Point = tmp;
			newDrawingInstruction.CorrespondingIndex = actIndexCorridorSegment;
			newDrawingInstruction.Angle = angleCorridorSorted[actIndexCorridorSegment];

			// x texture coordinate is taken from corridor segment point
			newDrawingInstruction.TextureXfromCorridor = true;
			newDrawingInstruction.Texture.X = profileCorridorSorted.TextureX[actIndexCorridorSegment];
			// y texture coordinate is computed by the distance to the corridor segment point
			newDrawingInstruction.Texture.Y = (newDrawingInstruction.Point - corridorVertices[actIndexCorridorSegment]).getLength() * distanceTextureYPerDistance1_;
			if (0 == end_)
				newDrawingInstruction.Texture.Y = -newDrawingInstruction.Texture.Y;
			else
				newDrawingInstruction.Texture.Y += lastTextureY_;

			// store drawing instruction
			adapterDrawingInstructions.push_back(newDrawingInstruction);

			// count on indices
			lastIndexCorridorSegment = actIndexCorridorSegment;
			++actIndexCorridorSegment;
			if (actIndexCorridorSegment == profileCorridorSorted.Point.size())
			{
				actIndexCorridorSegment = 0;
				closedCorridorSegment = true;
			}

		}
		else if (closedCorridorSegment || (anglesDockingSiteSorted[actIndexDockingSite]<angleCorridorSorted[actIndexCorridorSegment]) )
		// docking site vertex
		{
			// create new drawing instruction
			SAdapterDrawingInstruction newDrawingInstruction;
			newDrawingInstruction.Point = dockingSiteSorted.Point[actIndexDockingSite];
			newDrawingInstruction.Angle = anglesDockingSiteSorted[actIndexDockingSite];

			// search index of the corridor segment point with most similar angle
			angle1 = abs(anglesDockingSiteSorted[actIndexDockingSite] - angleCorridorSorted[actIndexCorridorSegment]);
			if (angle1>M_PI) angle1 = 2*M_PI - angle1; // smaller angle of 2PI
			angle2 = abs(anglesDockingSiteSorted[actIndexDockingSite] - angleCorridorSorted[lastIndexCorridorSegment]);
			if (angle2>M_PI) angle2 = 2*M_PI - angle2; // smaller angle of 2PI
			// index is the corresponding index of the drawing instruction
			if (angle1<angle2)
				newDrawingInstruction.CorrespondingIndex = actIndexCorridorSegment;
			else
				newDrawingInstruction.CorrespondingIndex = lastIndexCorridorSegment;

			// x texture coordinate will be computed later
			newDrawingInstruction.TextureXfromCorridor = false;
			
			// compute y texture coordinate:
			// compute interpolation factor
			angle1 =  newDrawingInstruction.Angle - angleCorridorSorted[lastIndexCorridorSegment];
			interpolationFactor = ComputePositionInterpolationFactor(
				profileCorridorSorted.Point[lastIndexCorridorSegment],
				profileCorridorSorted.Point[actIndexCorridorSegment],
				angle1);
			// compute coordinates of a temporary corridor segment point with similar angle
			tmp = (1-interpolationFactor) * corridorVertices[lastIndexCorridorSegment]
					+ interpolationFactor * corridorVertices[actIndexCorridorSegment];

			// y texture coordinate is computed by the distance to the temporary corridor segment point
			newDrawingInstruction.Texture.Y = (newDrawingInstruction.Point - tmp).getLength() * distanceTextureYPerDistance1_;
			if (0 == end_)
				newDrawingInstruction.Texture.Y = -newDrawingInstruction.Texture.Y;
			else
				newDrawingInstruction.Texture.Y += lastTextureY_;

			// store drawing instruction
			adapterDrawingInstructions.push_back(newDrawingInstruction);

			// count on indices
			lastIndexDockingSite = actIndexDockingSite;
			++actIndexDockingSite;
			if (actIndexDockingSite == dockingSiteSorted.Point.size())
			{
				actIndexDockingSite = 0;
				closedDockingSite = true;
			}
		}
		else
		// similar angles
		{
			// create new drawing instruction
			SAdapterDrawingInstruction newDrawingInstruction;
			newDrawingInstruction.Point = dockingSiteSorted.Point[actIndexDockingSite];
			newDrawingInstruction.Angle = anglesDockingSiteSorted[actIndexDockingSite];
			newDrawingInstruction.CorrespondingIndex = actIndexCorridorSegment;
			
			// x texture coordinate is taken from corridor segment point
			newDrawingInstruction.TextureXfromCorridor = true;
			newDrawingInstruction.Texture.X = profileCorridorSorted.TextureX[actIndexCorridorSegment];
			/// y texture coordinate is computed by the distance to the corridor segment point
			newDrawingInstruction.Texture.Y = (newDrawingInstruction.Point - corridorVertices[actIndexCorridorSegment]).getLength() * distanceTextureYPerDistance1_;
			if (end_ == 0)
				newDrawingInstruction.Texture.Y = -newDrawingInstruction.Texture.Y;
			else
				newDrawingInstruction.Texture.Y += lastTextureY_;

			// store drawing instruction
			adapterDrawingInstructions.push_back(newDrawingInstruction);

			// count on indices
			lastIndexCorridorSegment = actIndexCorridorSegment;
			++actIndexCorridorSegment;
			if (actIndexCorridorSegment == profileCorridorSorted.Point.size())
			{
				actIndexCorridorSegment = 0;
				closedCorridorSegment = true;
			}
			// count on indices
			lastIndexDockingSite = actIndexDockingSite;
			++actIndexDockingSite;
			if (actIndexDockingSite == dockingSiteSorted.Point.size())
			{
				actIndexDockingSite = 0;
				closedDockingSite = true;
			}
		}

	} // END while

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// compute x texture coordinates for drawing instructions
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	// if not TextureXfromCooridor x texture coordinate is interpolated
	// from the enclosing TextureXfromCorridor points,
	// interpolation is based on the distances between the docking-site-plane-projected points

	double fullLength, actLength;
	unsigned int interpolationIndex;

	for (unsigned int i = 0; i<adapterDrawingInstructions.size(); ++i)
		if (adapterDrawingInstructions[i].TextureXfromCorridor)
		{
			// compute full length of the walk between the 2 enclosing TextureXfromCorridors
			fullLength = 0;
			lastIndexDockingSite = i;
			actIndexDockingSite = i+1;
			if (actIndexDockingSite>=adapterDrawingInstructions.size())
				actIndexDockingSite=0;

			// search next TextureXfromCorridor
			while (!adapterDrawingInstructions[actIndexDockingSite].TextureXfromCorridor)
			{
				fullLength += ProjectVectorToPlane(adapterDrawingInstructions[actIndexDockingSite].Point-adapterDrawingInstructions[lastIndexDockingSite].Point, leftOfDS, upOfDS).getLength();
				lastIndexDockingSite = actIndexDockingSite;
				++actIndexDockingSite;
				if (actIndexDockingSite>=adapterDrawingInstructions.size())
					actIndexDockingSite=0;
			}
			fullLength += ProjectVectorToPlane(adapterDrawingInstructions[actIndexDockingSite].Point-adapterDrawingInstructions[lastIndexDockingSite].Point, leftOfDS, upOfDS).getLength();

			// prepare interpolation
			interpolationIndex = actIndexDockingSite;
			actLength = 0;
			lastIndexDockingSite = i;
			actIndexDockingSite = i+1;
			if (actIndexDockingSite>=adapterDrawingInstructions.size())
				actIndexDockingSite=0;

			// interpolate for all points between the enclosing TextureXfromCorridors
			while (!adapterDrawingInstructions[actIndexDockingSite].TextureXfromCorridor)
			{
				actLength += ProjectVectorToPlane(adapterDrawingInstructions[actIndexDockingSite].Point-adapterDrawingInstructions[lastIndexDockingSite].Point, leftOfDS, upOfDS).getLength();
				
				// interpolation factor is based on distance ratio
				interpolationFactor = actLength/fullLength;
				adapterDrawingInstructions[actIndexDockingSite].Texture.X = (1-interpolationFactor) * adapterDrawingInstructions[i].Texture.X
						 + interpolationFactor * adapterDrawingInstructions[interpolationIndex].Texture.X;

				lastIndexDockingSite = actIndexDockingSite;
				++actIndexDockingSite;
				if (actIndexDockingSite>=adapterDrawingInstructions.size())
					actIndexDockingSite=0;
			}
		};
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// draw the adapter
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	DrawAdapter(end_, lastTextureY_, profileCorridorSorted, corridorVertices, angleCorridorSorted, dockingSite_, adapterDrawingInstructions);
}

void DunGen::CCorridor::DrawAdapter(unsigned int end_,
	double lastTextureY_, const SCorridorProfile& profileCorridorSorted_,
	const std::vector<irr::core::vector3d<double>>& corridorVertices_, const std::vector<double>& angleCorridorSorted_,
	const SDockingSite& dockingSite_, const std::vector<SAdapterDrawingInstruction>& drawingInstructions_ )
{
	// create mesh buffer (only one needed per adapter)
	irr::scene::SMeshBuffer* meshBuffer = new irr::scene::SMeshBuffer();
	MeshAdapter[end_]->addMeshBuffer(meshBuffer);
	meshBuffer->drop();
	// reserve for worst case
	meshBuffer->Vertices.set_used(MaxVertexCount); 
	meshBuffer->Indices.set_used(3*MaxVertexCount);	// 3 indices per vertex (1 triangle per vertex)
	unsigned int bufferVertices = 0;
	unsigned int bufferIndizes = 0;

	// add corridor segment vertices:
	
	irr::core::vector3d<double> tmp;
	// for every point of the profile:
	for (unsigned int i = 0; i<corridorVertices_.size(); ++i)
	{
		irr::video::S3DVertex& v = meshBuffer->Vertices[bufferVertices++];
		v.Pos.set(vec3F(corridorVertices_[i]));

		// compute normal
		tmp = (Position[end_] - corridorVertices_[i]).normalize();
		v.Normal.set(vec3F(tmp));

		// set texture coordinates
		v.TCoords.set(static_cast<irr::f32>(profileCorridorSorted_.TextureX[i]),static_cast<irr::f32>(lastTextureY_));
	}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// creating triangles and docking site vertices
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	unsigned int lastCorrespondingIndex = drawingInstructions_[0].CorrespondingIndex;

	// create new vertex
	irr::video::S3DVertex& v = meshBuffer->Vertices[bufferVertices++];
	tmp = drawingInstructions_[0].Point;
	v.Pos.set(vec3F(tmp));

	// compute normal
	tmp = dockingSite_.Center - tmp;
	// projection on plane is used (better result in most cases)
	tmp -= dockingSite_.Normal.dotProduct(tmp)*dockingSite_.Normal;
	tmp.normalize();
	v.Normal.set(vec3F(tmp));

	// set texture coordinates
	v.TCoords.set(static_cast<irr::f32>(drawingInstructions_[0].Texture.X), static_cast<irr::f32>(drawingInstructions_[0].Texture.Y));

	double angle1, angle2;
	// remaining vertices
	for (unsigned int i = 1; i<drawingInstructions_.size(); ++i)
	{
		// create new vertex
		irr::video::S3DVertex& v = meshBuffer->Vertices[bufferVertices++];
		tmp = drawingInstructions_[i].Point;
		v.Pos.set(vec3F(tmp));

		// compute normal
		tmp = dockingSite_.Center - tmp;
		// projection on plane is used (better result in most cases)
		tmp -= dockingSite_.Normal.dotProduct(tmp)*dockingSite_.Normal;

		tmp.normalize();
		v.Normal.set(vec3F(tmp));

		// set texture coordinates
		v.TCoords.set(static_cast<irr::f32>(drawingInstructions_[i].Texture.X),static_cast<irr::f32>(drawingInstructions_[i].Texture.Y));

		// add 1 triangle?
		if (drawingInstructions_[i].CorrespondingIndex == lastCorrespondingIndex)
		{
			meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;

			if (0 == end_)	// direction is based on the end of the corridor
			{
				meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[i].CorrespondingIndex;
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-2;
			}
			else
			{
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-2;
				meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[i].CorrespondingIndex;
			}
		}
		// add 2 triangles?
		else		
		{
			// which of both vertices is nearer to the corresponding vertex (angle wise)
			// -> this vertex is on the diagonal axis of both triangles

			angle1 = abs(drawingInstructions_[i].Angle - angleCorridorSorted_[lastCorrespondingIndex]);
			if (angle1>M_PI) angle1 = 2*M_PI - angle1;	// smaller angle of 2PI
			angle2 = abs(drawingInstructions_[i-1].Angle  - angleCorridorSorted_[drawingInstructions_[i].CorrespondingIndex]);
			if (angle2>M_PI) angle2 = 2*M_PI - angle2;  // smaller angle of 2PI

			if (angle1<angle2)
			{
				// i and lastCorrespondingIndex are on the diagonal axis

				// add triangle:
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;
				if (0 == end_)	// direction is based on the end of the corridor
				{
					meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
					meshBuffer->Indices[bufferIndizes++] = bufferVertices-2;
				}
				else
				{
					meshBuffer->Indices[bufferIndizes++] = bufferVertices-2;
					meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
				}
				// add triangle:
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;
				if (0 == end_)	// direction is based on the end of the corridor
				{
					meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[i].CorrespondingIndex;
					meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
				}
				else
				{
					meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
					meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[i].CorrespondingIndex;
				}
			}
			else
			{
				// i-1 and the actual corresponding index are on the diagonal axis

				// add triangle:
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-2;
				if (0 == end_)	// direction is based on the end of the corridor
				{
					meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;
					meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[i].CorrespondingIndex;					
				}
				else
				{					
					meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[i].CorrespondingIndex;
					meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;
				}
				// add triangle:
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-2;
				if (0 == end_)	// direction is based on the end of the corridor
				{
					meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[i].CorrespondingIndex;
					meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
				}
				else
				{
					meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
					meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[i].CorrespondingIndex;
				}
			}

		} // END: inset triangles

		lastCorrespondingIndex = drawingInstructions_[i].CorrespondingIndex;

	} // END for

	// ----------------------------------
	// last triangles of the adapter
	// ----------------------------------

	// add 1 triangle?
	if (drawingInstructions_[0].CorrespondingIndex == lastCorrespondingIndex)
	{
		meshBuffer->Indices[bufferIndizes++] = bufferVertices-drawingInstructions_.size();

		if (0 == end_)	// direction is based on the end of the corridor
		{
			meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[0].CorrespondingIndex;
			meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;
		}
		else
		{
			meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;
			meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[0].CorrespondingIndex;
		}
	}
	// add 2 triangles?
	else		
	{
		// which of both vertices is nearer to the corresponding vertex (angle wise)
		// -> this vertex is on the diagonal axis of both triangles

		angle1 = abs(drawingInstructions_[0].Angle - angleCorridorSorted_[lastCorrespondingIndex]);
		if (angle1>M_PI) angle1 = 2*M_PI - angle1;	// smaller angle of 2PI
		angle2 = abs(drawingInstructions_.back().Angle  - angleCorridorSorted_[drawingInstructions_[0].CorrespondingIndex]);
		if (angle2>M_PI) angle2 = 2*M_PI - angle2;  // smaller angle of 2PI

		if (angle1<angle2)
		{
			// add triangle:
			meshBuffer->Indices[bufferIndizes++] = bufferVertices-drawingInstructions_.size(); // first docking site vertex

			if (0 == end_)	// direction is based on the end of the corridor
			{
				meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;
			}
			else
			{
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-1;
				meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
			}

			// add triangle:
			meshBuffer->Indices[bufferIndizes++] = bufferVertices-drawingInstructions_.size(); // first docking site vertex

			if (0 == end_)	// direction is based on the end of the corridor
			{
				meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[0].CorrespondingIndex;
				meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
			}
			else
			{
				meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
				meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[0].CorrespondingIndex;
			}
		}
		else
		{
			// add triangle:
			meshBuffer->Indices[bufferIndizes++] = bufferVertices-1; // last docking site vertex

			if (0 == end_)	// direction is based on the end of the corridor
			{
				meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[0].CorrespondingIndex;
				meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
			}
			else
			{				
				meshBuffer->Indices[bufferIndizes++] = lastCorrespondingIndex;
				meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[0].CorrespondingIndex;
			}

			// add triangle:
			meshBuffer->Indices[bufferIndizes++] = bufferVertices-1; // last docking site vertex

			if (0 == end_)	// direction is based on the end of the corridor
			{
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-drawingInstructions_.size();
				meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[0].CorrespondingIndex;

			}
			else
			{
				meshBuffer->Indices[bufferIndizes++] = drawingInstructions_[0].CorrespondingIndex;
				meshBuffer->Indices[bufferIndizes++] = bufferVertices-drawingInstructions_.size();
			}
		}

	} // END: inset triangles

	// compute final values of the meshbuffer
	meshBuffer->Vertices.reallocate(bufferVertices);
	meshBuffer->Indices.reallocate(bufferIndizes);
	meshBuffer->recalculateBoundingBox();

	// set final bounding box for the mesh
	MeshAdapter[end_]->recalculateBoundingBox();	
}