// Copyright (C) 2011-2013 by Maximilian Hönig
// This file is part of "DunGen - the Dungeongenerator".
// For conditions of distribution and use, see licence.txt provided together with DunGen.

#ifndef CORRIDOR_H
#define CORRIDOR_H

#include "interface/CorridorCommon.h"
#include <irrlicht.h>
#include <vector>

// Namespace DunGen : DungeonGenerator
namespace DunGen
{
	class CRandomGenerator;
	struct SDockingSite;

	/// a corridor in a dungeon
	///
	/// uses a cubical hermite spline as base curve
	class CCorridor
	{
	public:
		/// a corridor profile
		///
		/// a chain of 3d points as a closed curve
		/// clockwise direction
		/// important: polar angle of the points has to be monotonic increasing circulary
		/// the 0 point has to be enclosed by the projected curve  (-> at least 3 points necessary)
		struct SCorridorProfile
		{
			std::vector<irr::core::vector2d<double>> Point;	///< points of the profile
			std::vector<double> TextureX;					///< x texture coordinate of each point
		};

		/// detail objects along the corridor: e.g. torches
		struct SDetailObject
		{
			irr::scene::ISceneNode* Node;							///< node of all instances
			irr::core::vector3d<double> Scale;						///< scale of all instances
			std::vector<irr::core::vector3d<double>> Position;		///< position of each instance
			std::vector<irr::core::vector3d<double>> Rotation;		///< rotation of each instance
		};

	private:
		/// instructions for drawing a adapter
		struct SAdapterDrawingInstruction
		{
			irr::core::vector3d<double> Point;		///< point to draw
			irr::core::vector2d<double> Texture;	///< texture coordiantes of the point
			double Angle;							///< polar angle of the point when projected in docking site plane
			unsigned int CorrespondingIndex;		///< which corridor segment index has most similar angle?
			bool TextureXfromCorridor;				///< is the x texture coordiante directly taken from the corridor segment?
		};

		/// view region for visibility test
		struct SViewRegionOfInterest
		{
			irr::core::vector3d<double> Position;	///< position
			irr::core::vector3d<double> Left;		///< left vector
			irr::core::vector3d<double> Up;			///< up vector
			irr::core::rect<double> ProfileBB;		///< rectangular profile of the region
		};

		/// viewing volume for visibility test
		struct SViewingVolume
		{
			/// constructor, takes 2 regions of interest to form a viewing volume
			SViewingVolume(const SViewRegionOfInterest* region1_, const SViewRegionOfInterest* region2_);

			irr::core::vector3d<double> PointOfReference;	///< center of the projection plane
			irr::core::vector3d<double> Left;				///< left vector spanning the projection plane
			irr::core::vector3d<double> Up;					///< up vector spanning the projection plane
			irr::core::rect<double> ViewingVolumeBB;		///< projected bounding box of the viewing volume
		};

	public:
		/// constructor: create corridor
		///
		/// important: derivation has to be unequal to the y axis vector (0,y,0)
		CCorridor(const SCorridorProfile& profile_, const SDockingSite& dockingSite0_, const SDockingSite& dockingSite1_,
			const irr::core::vector3d<double>& position0_, const irr::core::vector3d<double>& position1_,
			const irr::core::vector3d<double>& derivation0_, const irr::core::vector3d<double>& derivation1_,
			double distance_, double distanceTextureYPerDistance1_);
		 /// destructor
		~CCorridor();

		/// place detailobject
		/// 
		/// rotation is the rotation, if the corridor points to (0,0,1)
		/// position in 2D to the center of a corridor segment
		/// object is placed in random distance DistanceSampling*DistanceNumFactor*[DistanceNumMin,DistanceNumMax]
		/// first element is placed in distance DistanceSampling*DistanceNumFactor*[DistanceNumMinFirstElement,DistanceNumMaxFirstElement] from position0
		/// if ObjectAtT1, an object is placed at position1, even if the distance to the last object would be too small
		///
		/// important: Max>=Min, DistanceNumMin>=1, DistanceNumMinFirstElement>=0
		void PlaceDetailObject(const SDetailobjectParameters& parameters_, const CRandomGenerator* randomGenerator_);

		/// is the corridor sight blocking: do one can not see from one end to the other?
		bool GetDefinitivelySightBlocking();

		/// read mesh corridor
		irr::scene::SMesh* GetMesh();

		/// read mesh adapter
		irr::scene::SMesh* GetMeshAdapter(unsigned int i_);

		/// reads a detailobject, if no detailobject with this index exits, NULL is returned
		SDetailObject* GetDetailObject(unsigned int i_);

	private:
		/// creates the triangle mesh of the corridor
		///
		/// _Distance ... desired distance between the segments of the corridor
		/// \returns lastTextureY
		double CreateCorridor(const SCorridorProfile& profile_, const SDockingSite& dockingSite0_, const SDockingSite& dockingSite1_, double distance_, double distanceTextureYPerDistance1_);

		/// create adapter, connects a docking site and the corridor
		///
		// _DistanceTextureYPerDistance1 ... normalized texture distance
		void CreateAdapter(unsigned int end_, const SDockingSite& dockingSite_, const SCorridorProfile& profile_, double lastTextureY_, double distanceTextureYPerDistance1_);

		/// creates the adapters vertices and triangles
		void DrawAdapter(unsigned int end_, double lastTextureY_,
			const SCorridorProfile& profileCorridorSorted_, const std::vector<irr::core::vector3d<double>>& corridorVertices_,
			const std::vector<double>& angleCorridorSorted_, const SDockingSite& dockingSite_,
			const std::vector<SAdapterDrawingInstruction>& drawingInstructions_);

		/// computes new t value for point in given destination
		///
		/// \returns (new t, distance from point of last t to point of new t)
		inline irr::core::vector2d<double> ComputeT(const irr::core::vector3d<double>& lastPosition_, double lastT_, double distance_);

		/// returns position to given t
		inline irr::core::vector3d<double> ComputePosition(double t_);
		/// returns derivation to given t
		inline irr::core::vector3d<double> ComputeDerivation(double t_);

		/// compute the interpolation factor for position P, P is on line segment AB of triangle AMB
		inline double ComputePositionInterpolationFactor(const irr::core::vector2d<double>& ma_, const irr::core::vector2d<double>& mb_, double angleAMP_);
		

		/// project a vector onto a plane spanned by given up and left vector (both normalized)
		///
		/// \returns (X,Y)-coordiantes in the plane with X+ ... -left, Y+ ... up
		inline irr::core::vector2d<double> ProjectVectorToPlane(const irr::core::vector3d<double>& vector_, const irr::core::vector3d<double>& left_, const irr::core::vector3d<double>& up_);

		/// compute cubical hermite spline coefficents for position and derivation
		void ComputeResultingCoefficients();

		/// test, if the corridor blocks sight
		void VisibilityTest(const SCorridorProfile& profile_, const SDockingSite& dockingSite0_, const SDockingSite& dockingSite1_);

		/// tests, if a region of interest is outside a viewing volume (if yes -> sighblocking)
		inline bool IsSightBlocking(const SViewingVolume* viewingVolume_, const SViewRegionOfInterest* regionOfInterest_);		

		/// triangular mesh of the corridor
		irr::scene::SMesh* MeshCorridor;
		/// triangular meshes of the adapters
		irr::scene::SMesh* MeshAdapter[2];

		/// detail objects
		std::vector<SDetailObject*> DetailObjects;

		/// start- and endposition of the corridor
		irr::core::vector3d<double> Position[2];
		/// derivation of the hermite curve in the start- and endposition of the corridor
		irr::core::vector3d<double> Derivation[2];

		/// coefficients for computing the position P(t) 3x4 matrix
		///
		/// P(t) = (t^3,t^2,t,1) * PositionCoefficients[0...3]
		irr::core::vector3d<double> PositionCoefficients[4];
		/// coefficients for computing the first derivation P1(t): 3x3 matrix
		///
		/// P1(t) = (t^2,t,1) * DerivationCoefficients[0...2]
		irr::core::vector3d<double> DerivationCoefficients[3];

		/// is the corridor sight blocking: do one can not see from one end to the other?
		bool DefinitivelySightBlocking;

		/// region of interests
		///
		/// number is the number of local maxima regarding the distance point<->axis P(1)-P(0) -> normally 1 or 2 values
		/// regions are ordered from P(0) to P(1)
		std::vector<SViewRegionOfInterest> ViewTestRegions;
		
		/// standard vector for Up, defines the height axis
		static const irr::core::vector3d<double> UpStandard;

		/// precision for distance when searching the next paramenter t
		static const double Precision;

		/// maximal number of vertices per meshbuffer
		static const unsigned int MaxVertexCount = 65500;
	};

} // END NAMESPACE DunGen

#endif