// Copyright (C) 2011-2013 by Maximilian Hönig
// This file is part of "DunGen - the Dungeongenerator".
// For conditions of distribution and use, see licence.txt provided together with DunGen.

#include "Corridor.h"
#include "DockingSite.h"
#include "RandomGenerator.h"

// ======================================================
// detail objects
// ======================================================

void DunGen::CCorridor::PlaceDetailObject(const SDetailobjectParameters& parameters_, const CRandomGenerator* randomGenerator_)
{
	// create new detail object
	SDetailObject* newDetailObject = new SDetailObject();
	newDetailObject->Node = parameters_.Node;
	newDetailObject->Scale = parameters_.Scale;
	
	// compute matrix of the basic rotation
	irr::core::CMatrix4<double> basicRotationMatrix(irr::core::CMatrix4<double>::EM4CONST_IDENTITY);
	basicRotationMatrix.setRotationDegrees(parameters_.Rotation);
	
	// matrix of the spline rotation
	irr::core::CMatrix4<double> splineRotationMatrix(irr::core::CMatrix4<double>::EM4CONST_NOTHING);

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// place objects along the spline
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	

	// multiple of the _DistanceSampling
	unsigned int distanceFactor = parameters_.DistanceNumFactor*
		randomGenerator_->GetRandomNumberMinMax(parameters_.DistanceNumMinFirstElement,parameters_.DistanceNumMaxFirstElement);
	// spline parameter t
	double t = 0.0;
	// position and  derivation for current t value
	irr::core::vector3d<double> actPosition = ComputePosition(t);
	irr::core::vector3d<double> actDerivation;
	// local coordinate system
	irr::core::vector3d<double> front;
	irr::core::vector3d<double> left;
	irr::core::vector3d<double> up;

	// if distance is 0 -> place object
	if (distanceFactor==0)
	{
		// compute coordinate system (left handed)
		actDerivation = ComputeDerivation(t);	
		front = actDerivation.normalize();
		left = front.crossProduct(UpStandard);
		left.normalize();
		up = left.crossProduct(front);

		// resulting position: (-left,up)*(X,Y)
		newDetailObject->Position.push_back(actPosition - parameters_.Position.X*left + parameters_.Position.Y*up);

		// compute angle between (0,0,1) and front
		// and setup rotation matrix
		splineRotationMatrix.makeIdentity();
		splineRotationMatrix.setRotationDegrees(front.getHorizontalAngle());
		// aggregate rotations
		newDetailObject->Rotation.push_back((splineRotationMatrix*basicRotationMatrix).getRotationDegrees());

		// compute new distance factor
		distanceFactor = parameters_.DistanceNumFactor*
			randomGenerator_->GetRandomNumberMinMax(parameters_.DistanceNumMin,parameters_.DistanceNumMax);
	}

	// go on further on the spline
	irr::core::vector2d<double> computeTResult;
	while (t < 1.0)
	{
		// actualize
		computeTResult = ComputeT(actPosition, t, parameters_.DistanceSampling);
		t = computeTResult.X;
		actPosition = ComputePosition(t);
		--distanceFactor;

		// if not enough distance (can be for the last t): break
		if ((computeTResult.Y < (parameters_.DistanceSampling-Precision)) && (!parameters_.ObjectAtT1))
			break;

		// if distance is 0 -> place object
		if ((distanceFactor==0)||(t==1.0))
		{
			// compute coordinate system (left handed)
			actDerivation = ComputeDerivation(t);
			front = actDerivation.normalize();
			left = front.crossProduct(UpStandard);
			left.normalize();
			up = left.crossProduct(front);

			// resulting position: (-left,up)*(X,Y)
			newDetailObject->Position.push_back(actPosition - parameters_.Position.X*left + parameters_.Position.Y*up);

			// compute angle between (0,0,1) and front
			// and setup rotation matrix
			splineRotationMatrix.makeIdentity();
			splineRotationMatrix.setRotationDegrees(front.getHorizontalAngle());
			// aggregate rotations
			newDetailObject->Rotation.push_back((splineRotationMatrix*basicRotationMatrix).getRotationDegrees());

			// compute new distance factor
			distanceFactor = parameters_.DistanceNumFactor*
				randomGenerator_->GetRandomNumberMinMax(parameters_.DistanceNumMin,parameters_.DistanceNumMax);
		}
	}
	
	// store detailobjects
	DetailObjects.push_back(newDetailObject);
}

// ======================================================
// visibility test
// ======================================================

inline bool DunGen::CCorridor::IsSightBlocking(const SViewingVolume* viewingVolume_, const SViewRegionOfInterest* regionOfInterest_)
{
	irr::core::vector2d<double> tmpVec1;
	irr::core::vector3d<double> tmpVec2;

	// roiBB ... projected bounding box of the region of interest
	irr::core::rect<double> roiBB;
		
	// project BB RegionOfInterest onto plane of the ViewingVolume
	tmpVec2 = -regionOfInterest_->ProfileBB.UpperLeftCorner.X * regionOfInterest_->Left + regionOfInterest_->ProfileBB.UpperLeftCorner.Y * regionOfInterest_->Up;	// as 3D-vector
	tmpVec2 += regionOfInterest_->Position - viewingVolume_->PointOfReference;
	tmpVec1.X = -tmpVec2.dotProduct(viewingVolume_->Left);	// projection
	tmpVec1.Y = tmpVec2.dotProduct(viewingVolume_->Up);
	roiBB.UpperLeftCorner = tmpVec1;	// first point of the roiBB
	roiBB.LowerRightCorner = tmpVec1;

	tmpVec2 = -regionOfInterest_->ProfileBB.LowerRightCorner.X * regionOfInterest_->Left + regionOfInterest_->ProfileBB.LowerRightCorner.Y * regionOfInterest_->Up;	// as 3D-vector
	tmpVec2 += regionOfInterest_->Position - viewingVolume_->PointOfReference;
	tmpVec1.X = -tmpVec2.dotProduct(viewingVolume_->Left);	// projection
	tmpVec1.Y = tmpVec2.dotProduct(viewingVolume_->Up);
	roiBB.addInternalPoint(tmpVec1);	// add point

	tmpVec2 = -regionOfInterest_->ProfileBB.UpperLeftCorner.X * regionOfInterest_->Left + regionOfInterest_->ProfileBB.LowerRightCorner.Y * regionOfInterest_->Up;	// as 3D-vector
	tmpVec2 += regionOfInterest_->Position - viewingVolume_->PointOfReference;
	tmpVec1.X = -tmpVec2.dotProduct(viewingVolume_->Left);	// projection
	tmpVec1.Y = tmpVec2.dotProduct(viewingVolume_->Up);
	roiBB.addInternalPoint(tmpVec1);	// add point

	tmpVec2 = -regionOfInterest_->ProfileBB.LowerRightCorner.X * regionOfInterest_->Left + regionOfInterest_->ProfileBB.UpperLeftCorner.Y * regionOfInterest_->Up;	// as 3D-vector
	tmpVec2 += regionOfInterest_->Position - viewingVolume_->PointOfReference;
	tmpVec1.X = -tmpVec2.dotProduct(viewingVolume_->Left);	// projection
	tmpVec1.Y = tmpVec2.dotProduct(viewingVolume_->Up);
	roiBB.addInternalPoint(tmpVec1);	// add point
	
	// test for intersection: if is not intersecting, no view ray can pass the viewing volume
	// so this would be sight blocking
	return (!roiBB.isRectCollided(viewingVolume_->ViewingVolumeBB));	
}

void DunGen::CCorridor::VisibilityTest(const SCorridorProfile& profile_, const SDockingSite& dockingSite0_, const SDockingSite& dockingSite1_)
{
	DefinitivelySightBlocking = false;

	// dimensions of the corridor profile
	irr::core::rect<double> corridorProfileBB;
	// regions of interests of docking sites
	SViewRegionOfInterest rOIDockingSite[2];
	
	// compute corridor profile dimensions
	corridorProfileBB.UpperLeftCorner.X = 0.0;
	corridorProfileBB.UpperLeftCorner.Y = 0.0;
	corridorProfileBB.LowerRightCorner.X = 0.0;
	corridorProfileBB.LowerRightCorner.Y = 0.0;
	for (unsigned int i=0; i<profile_.Point.size(); i++)
	{
		// upper left: smaller values
		if (profile_.Point[i].X < corridorProfileBB.UpperLeftCorner.X)
			corridorProfileBB.UpperLeftCorner.X = profile_.Point[i].X;
		if (profile_.Point[i].Y < corridorProfileBB.UpperLeftCorner.Y)
			corridorProfileBB.UpperLeftCorner.Y = profile_.Point[i].Y;
		
		// lower right: larger values
		if (profile_.Point[i].X > corridorProfileBB.LowerRightCorner.X)
			corridorProfileBB.LowerRightCorner.X = profile_.Point[i].X;
		if (profile_.Point[i].Y > corridorProfileBB.LowerRightCorner.Y)
			corridorProfileBB.LowerRightCorner.Y = profile_.Point[i].Y;
	}

	// transfer spline profile BB to the already computed view test regions
	for (unsigned int i=0; i<ViewTestRegions.size(); ++i)
		ViewTestRegions[i].ProfileBB = corridorProfileBB;

	// compute docking site ROIs
	irr::core::vector3d<double> front;
	const SDockingSite* dockingSite[2];
	dockingSite[0] = &dockingSite0_;
	dockingSite[1] = &dockingSite1_;
	irr::core::vector2d<double> tmpVec2D;		

	for (unsigned int i=0; i<2; i++)
	{
		front = dockingSite[i]->Normal;
		rOIDockingSite[i].Left = (front.crossProduct(UpStandard)).normalize();
		rOIDockingSite[i].Up = (rOIDockingSite[i].Left).crossProduct(front);
		// position is at the inner end of the docking site
		if (0==i)
			rOIDockingSite[i].Position = dockingSite[i]->Center + dockingSite[i]->Extend*dockingSite[i]->Normal;
		else
			rOIDockingSite[i].Position = dockingSite[i]->Center - dockingSite[i]->Extend*dockingSite[i]->Normal;
		
		rOIDockingSite[i].ProfileBB.UpperLeftCorner.X = 0.0;
		rOIDockingSite[i].ProfileBB.UpperLeftCorner.Y = 0.0;
		rOIDockingSite[i].ProfileBB.LowerRightCorner.X = 0.0;
		rOIDockingSite[i].ProfileBB.LowerRightCorner.Y = 0.0;

		for (unsigned int j=0; j<dockingSite[i]->Point.size(); ++j)
		{
			// compute dimension: points have to be projected
			tmpVec2D = ProjectVectorToPlane(dockingSite[i]->Point[j] - dockingSite[i]->Center, rOIDockingSite[i].Left, rOIDockingSite[i].Up);

			// upper left: smaller values
			if (tmpVec2D.X < rOIDockingSite[i].ProfileBB.UpperLeftCorner.X)
				rOIDockingSite[i].ProfileBB.UpperLeftCorner.X = tmpVec2D.X;

			if (tmpVec2D.Y < rOIDockingSite[i].ProfileBB.UpperLeftCorner.Y)
				rOIDockingSite[i].ProfileBB.UpperLeftCorner.Y = tmpVec2D.Y;
		
			// lower right: larger values
			if (tmpVec2D.X > rOIDockingSite[i].ProfileBB.LowerRightCorner.X)
				rOIDockingSite[i].ProfileBB.LowerRightCorner.X = tmpVec2D.X;

			if (tmpVec2D.Y > rOIDockingSite[i].ProfileBB.LowerRightCorner.Y)
				rOIDockingSite[i].ProfileBB.LowerRightCorner.Y = tmpVec2D.Y;		
		}

		// because adapter use both, the corridor profile and the docking site: merge both BB
		rOIDockingSite[i].ProfileBB.addInternalPoint(corridorProfileBB.LowerRightCorner);
		rOIDockingSite[i].ProfileBB.addInternalPoint(corridorProfileBB.UpperLeftCorner);
	}

	// aggregate viewtest regions
	std::vector<SViewRegionOfInterest*> viewTestRegionsComplete;
	viewTestRegionsComplete.push_back(&rOIDockingSite[0]);
	for (unsigned int i=0; i<ViewTestRegions.size(); ++i)
		viewTestRegionsComplete.push_back(&ViewTestRegions[i]);
	viewTestRegionsComplete.push_back(&rOIDockingSite[1]);

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// create viewing volumes
	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	std::vector<SViewingVolume*> viewingVolumes;
	// comparing indices (all test objects have to be between both ends of the viewing volume)
	std::vector<unsigned int> testIndexStart;
	std::vector<unsigned int> testIndexEnd;
	for (unsigned int i=0; i<viewTestRegionsComplete.size(); ++i)
		for (unsigned int j=i+2; j<viewTestRegionsComplete.size(); ++j)	// minimum one region between
		{
			viewingVolumes.push_back(new SViewingVolume(viewTestRegionsComplete[i], viewTestRegionsComplete[j]) );
			// all regions between the "parent" regions have to be tested
			testIndexStart.push_back(i+1);
			testIndexEnd.push_back(j-1);
		}

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// test
	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~	

	for (unsigned int i=0; i<viewingVolumes.size(); ++i)
		for (unsigned int j=testIndexStart[i]; j<=testIndexEnd[i]; ++j)
			DefinitivelySightBlocking = DefinitivelySightBlocking || IsSightBlocking(viewingVolumes[i],viewTestRegionsComplete[j]);

	// delete the temporally created objects
	for (unsigned int i=0; i<viewingVolumes.size(); ++i)
		delete viewingVolumes[i];
}