// Copyright (C) 2011-2013 by Maximilian Hönig
// This file is part of "DunGen - the Dungeongenerator".
// For conditions of distribution and use, see licence.txt provided together with DunGen.

#include "Corridor.h"
#include "DockingSite.h"

// ======================================================
// structures
// ======================================================

DunGen::CCorridor::SViewingVolume::SViewingVolume(const SViewRegionOfInterest* region1_, const SViewRegionOfInterest* region2_)
{
	// center of the projection plane
	PointOfReference = region1_->Position;
	// vectors to span the projection plane
	irr::core::vector3d<double> tmpVec1 = (region2_->Position - region1_->Position).normalize();	// front
	Left = (tmpVec1.crossProduct(CCorridor::UpStandard)).normalize();
	Up = Left.crossProduct(tmpVec1);	
	irr::core::vector2d<double> tmpVec2;

	// BB (bounding box) 1 projection
	// project all 4 corner points and add then to the viewing volume BB
	tmpVec1 = -region1_->ProfileBB.UpperLeftCorner.X * region1_->Left + region1_->ProfileBB.UpperLeftCorner.Y * region1_->Up;	// as 3D-vector
	tmpVec2.X = -tmpVec1.dotProduct(Left);	// projection
	tmpVec2.Y = tmpVec1.dotProduct(Up);
	ViewingVolumeBB.UpperLeftCorner = tmpVec2; // first point of the VV-BB
	ViewingVolumeBB.LowerRightCorner = tmpVec2;

	tmpVec1 = -region1_->ProfileBB.LowerRightCorner.X * region1_->Left + region1_->ProfileBB.LowerRightCorner.Y * region1_->Up;	// as 3D-vector
	tmpVec2.X = -tmpVec1.dotProduct(Left);	// projection
	tmpVec2.Y = tmpVec1.dotProduct(Up);
	ViewingVolumeBB.addInternalPoint(tmpVec2);	// add point

	tmpVec1 = -region1_->ProfileBB.UpperLeftCorner.X * region1_->Left + region1_->ProfileBB.LowerRightCorner.Y * region1_->Up;	// as 3D-vector
	tmpVec2.X = -tmpVec1.dotProduct(Left);	// projection
	tmpVec2.Y = tmpVec1.dotProduct(Up);
	ViewingVolumeBB.addInternalPoint(tmpVec2);	// add point

	tmpVec1 = -region1_->ProfileBB.LowerRightCorner.X * region1_->Left + region1_->ProfileBB.UpperLeftCorner.Y * region1_->Up;	// as 3D-vector
	tmpVec2.X = -tmpVec1.dotProduct(Left);	// projection
	tmpVec2.Y = tmpVec1.dotProduct(Up);
	ViewingVolumeBB.addInternalPoint(tmpVec2);	// add point
	
	// BB 2 projection
	// project all 4 corner points and add then to the viewing volume BB
	tmpVec1 = -region2_->ProfileBB.UpperLeftCorner.X * region2_->Left + region2_->ProfileBB.UpperLeftCorner.Y * region2_->Up;	// as 3D-vector
	tmpVec2.X = -tmpVec1.dotProduct(Left);	// projection
	tmpVec2.Y = tmpVec1.dotProduct(Up);
	ViewingVolumeBB.addInternalPoint(tmpVec2);	// add point

	tmpVec1 = -region2_->ProfileBB.LowerRightCorner.X * region2_->Left + region2_->ProfileBB.LowerRightCorner.Y * region2_->Up;	// as 3D-vector
	tmpVec2.X = -tmpVec1.dotProduct(Left);	// projection
	tmpVec2.Y = tmpVec1.dotProduct(Up);
	ViewingVolumeBB.addInternalPoint(tmpVec2);	// add point

	tmpVec1 = -region2_->ProfileBB.UpperLeftCorner.X * region2_->Left + region2_->ProfileBB.LowerRightCorner.Y * region2_->Up;	// as 3D-vector
	tmpVec2.X = -tmpVec1.dotProduct(Left);	// projection
	tmpVec2.Y = tmpVec1.dotProduct(Up);
	ViewingVolumeBB.addInternalPoint(tmpVec2);	// add point

	tmpVec1 = -region2_->ProfileBB.LowerRightCorner.X * region2_->Left + region2_->ProfileBB.UpperLeftCorner.Y * region2_->Up;	// as 3D-vector
	tmpVec2.X = -tmpVec1.dotProduct(Left);	// projection
	tmpVec2.Y = tmpVec1.dotProduct(Up);
	ViewingVolumeBB.addInternalPoint(tmpVec2);	/// add point
}

// ======================================================
// initialization & de-initialization & read values
// ======================================================

// set precision
const double DunGen::CCorridor::Precision = 0.00001;

// set height axis (Y axis here)
const irr::core::vector3d<double> DunGen::CCorridor::UpStandard = irr::core::vector3d<double>(0,1,0);

DunGen::CCorridor::CCorridor(const SCorridorProfile& profile_, const SDockingSite& dockingSite0_, const SDockingSite& dockingSite1_,
	const irr::core::vector3d<double>& position0_, const irr::core::vector3d<double>& position1_,
	const irr::core::vector3d<double>& derivation0_, const irr::core::vector3d<double>& derivation1_,
	double distance_, double distanceTextureYPerDistance1_)
{
	// store values
	Position[0] = position0_;
	Position[1] = position1_;
	Derivation[0] = derivation0_;
	Derivation[1] = -derivation1_; // reverse derivation, so it is now in correct spline direction
	// compute the spline coefficients
	ComputeResultingCoefficients();
		
	// reverse order of docking site 2
	SDockingSite dockingSite2;
	for (unsigned int i=1; i<=dockingSite1_.Point.size(); ++i)
		dockingSite2.Point.push_back(dockingSite1_.Point[dockingSite1_.Point.size()-i]);
	dockingSite2.Normal = -dockingSite1_.Normal;
	dockingSite2.Center = dockingSite1_.Center;
	dockingSite2.Extend = dockingSite1_.Extend;
	
	// create the corridor
	double lastTextureY = CreateCorridor(profile_,dockingSite0_,dockingSite2,distance_,distanceTextureYPerDistance1_);

	// create adapters
	CreateAdapter(0,dockingSite0_,profile_, 0, distanceTextureYPerDistance1_);
	CreateAdapter(1,dockingSite2,profile_, lastTextureY, distanceTextureYPerDistance1_);

	// test for visibility
	VisibilityTest(profile_, dockingSite0_, dockingSite2);
}

DunGen::CCorridor::~CCorridor()
{
	// drop meshes
	MeshCorridor->drop();
	MeshAdapter[0]->drop();
	MeshAdapter[1]->drop();

	// delete detail objects
	for (unsigned int i=0; i<DetailObjects.size(); ++i)
		delete DetailObjects[i];
}

void DunGen::CCorridor::ComputeResultingCoefficients()
{
	// compute position coefficients
	PositionCoefficients[0] = 2 * Position[0] - 2 * Position[1] + Derivation[0] + Derivation[1];
	PositionCoefficients[1] = -3 * Position[0] + 3 * Position[1] - 2 * Derivation[0] - Derivation[1];
	PositionCoefficients[2] = Derivation[0];
	PositionCoefficients[3] = Position[0];

	// compute derivation coefficients
	DerivationCoefficients[0] = 6 * Position[0] -6 * Position[1] + 3 * Derivation[0] + 3 * Derivation[1];
	DerivationCoefficients[1] = -6 * Position[0] + 6 * Position[1] - 4 * Derivation[0] - 2 * Derivation[1];
	DerivationCoefficients[2] = Derivation[0];
}

bool DunGen::CCorridor::GetDefinitivelySightBlocking()
{
	return DefinitivelySightBlocking;
}

irr::scene::SMesh* DunGen::CCorridor::GetMesh()
{
	return MeshCorridor;
}

irr::scene::SMesh* DunGen::CCorridor::GetMeshAdapter(unsigned int i_)
{
	return MeshAdapter[i_];
}

DunGen::CCorridor::SDetailObject* DunGen::CCorridor::GetDetailObject(unsigned int i_)
{
	if (DetailObjects.size() <= i_)
		return NULL;
	return DetailObjects[i_];
}
